@@ -144,6 +144,7 @@ class TamingDecoder(ImageGenerator):
         loss_type: str = 'cosine_similarity',
         loss_clip_value: float = None,
         generation_cb=None,
+        init_embed: torch.Tensor = None,
         **kwargs,
     ) -> Tuple[List[PIL.Image.Image], List[torch.Tensor]]:
         """
@@ -172,6 +173,8 @@ class TamingDecoder(ImageGenerator):
                 inputs of this function will be the generated image
                 from the VQGAN (Nx3xHxW) and the logits used to 
                 generate it (Nx256xDxD) and the optimization step.
+            init_embed (torh.Tensor, optional): initial embedding 
+                From where to start the generation. Defaults to None.
 
         Returns:
             Tuple[List[PIL.Image.Image], List[torch.Tensor]]: list 
@@ -197,10 +200,10 @@ class TamingDecoder(ImageGenerator):
             embed_size,
         ).to(self.device)
 
-        if init_img_path is None:
-            z_logits = torch.nn.Parameter(
-                torch.sinh(1.9 * torch.arcsinh(z_logits)), )
-        else:
+        if init_embed is not None:
+            z_logits = torch.nn.Parameter(init_embed.to(self.device))
+
+        elif init_img_path is not None:
             pil_img = Image.open(init_img_path)
             pil_img = pil_img.convert('RGB')
 
@@ -217,6 +220,10 @@ class TamingDecoder(ImageGenerator):
             z_logits = torch.nn.Parameter(z)
             # img_z_logits = z.detach().clone()
 
+        else:
+            z_logits = torch.nn.Parameter(
+                torch.sinh(1.9 * torch.arcsinh(z_logits)), )
+
         torch.nn.Parameter(torch.ones((4, )))
 
         optimizer = torch.optim.AdamW(
@@ -325,6 +332,100 @@ class TamingDecoder(ImageGenerator):
 
         return gen_img_list
 
+    def zoom(
+        self,
+        init_latent: torch.Tensor,
+        lr: float = 0.008,
+        num_generations: int = 100000,
+        num_zoom_interp_steps=2,
+        num_zoom_train_steps=8,
+        zoom_offset=4,
+    ):
+        z_logits = init_latent.detach().clone()
+        z_logits = torch.nn.Parameter(z_logits)
+
+        optimizer = torch.optim.AdamW(
+            params=[z_logits],
+            lr=lr,
+            betas=(0.9, 0.999),
+            weight_decay=0.1,
+        )
+
+        gen_img_list = []
+        z_logits_list = []
+        for step in range(num_generations):
+            logging.info(f"Generation {step}/{num_generations}")
+
+            with torch.no_grad():
+                z = self.vqgan_model.post_quant_conv(z_logits)
+                x_rec = self.vqgan_model.decoder(z)
+                x_rec = (x_rec.clip(-1, 1) + 1) / 2
+
+                x_rec_size = x_rec.shape[-1]
+
+                x_rec_zoom = x_rec[:, :, zoom_offset:-zoom_offset,
+                                   zoom_offset:-zoom_offset]
+                x_rec_zoom = torch.nn.functional.interpolate(
+                    x_rec_zoom,
+                    (x_rec_size, x_rec_size),
+                    mode="bilinear",
+                )
+
+                x_rec_zoom = 2. * x_rec_zoom - 1
+                zoom_z_logits, _, [_, _, indices
+                                   ] = self.vqgan_model.encode(x_rec_zoom)
+
+                z_logits.data = zoom_z_logits.clone().detach()
+
+            for zoom_train_step in range(num_zoom_train_steps):
+                loss = 0
+                z = self.vqgan_model.post_quant_conv(z_logits)
+                x_rec = self.vqgan_model.decoder(z)
+                x_rec = (x_rec.clip(-1, 1) + 1) / 2
+                x_rec_stacked = self.augment(
+                    x_rec,
+                    x_rec.shape[1],
+                    x_rec.shape[2],
+                )
+
+                loss += 10 * self.compute_clip_loss(x_rec_stacked, prompt)
+
+                logging.info(
+                    f"Loss {loss} - {zoom_train_step}/{num_zoom_train_steps}")
+
+                optimizer.zero_grad()
+                loss.backward()
+                optimizer.step()
+
+            z_logits_1 = init_latent
+            z_logits_2 = z_logits
+
+            for zoom_step in range(num_zoom_interp_steps):
+                weight = zoom_step / num_zoom_interp_steps
+                interp_logits = weight * z_logits_2 + (1 - weight) * z_logits_1
+
+                with torch.no_grad():
+                    z = self.vqgan_model.post_quant_conv(interp_logits)
+                    x_rec = self.vqgan_model.decoder(z)
+                    x_rec = (x_rec.clip(-1, 1) + 1) / 2
+
+                x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(
+                    x_rec[0])
+                gen_img_list.append(x_rec_img)
+
+                print("Adding img...")
+                os.makedirs("generations", exist_ok=True)
+                x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(
+                    x_rec[0])
+                x_rec_img.save(
+                    f"generations/{step}_{zoom_step}_{zoom_train_step}.jpg")
+
+            init_latent = z_logits.detach().clone()
+
+            torch.cuda.empty_cache()
+
+        return gen_img_list, z_logits_list
+
 
 if __name__ == '__main__':
     target_img_size = 256
@@ -337,26 +438,37 @@ if __name__ == '__main__':
     # init_img_path = "alien.png"
     loss_clip_value = None
 
-    def save_img_cb(
-        x_rec,
-        _z_logits,
-        step,
-    ):
-        x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(x_rec[0])
-        x_rec_img.save(f"{step}.jpg")
-
     taming_decoder = TamingDecoder()
-    gen_img_list, z_logits_list = taming_decoder.generate_from_prompt(
-        prompt=prompt,
-        lr=lr,
-        target_img_size=target_img_size,
-        num_steps=num_steps,
-        num_augmentations=num_augmentations,
-        init_img_path=init_img_path,
-        loss_type=loss_type,
-        loss_clip_value=loss_clip_value,
-        generation_cb=save_img_cb,
-    )
+
+    init_latent_path = f"./{prompt}_logits.pt"
+    if os.path.exists(init_latent_path):
+        init_latent = torch.load(init_latent_path).to('cuda')
+    else:
+
+        def save_img_cb(
+            x_rec,
+            _z_logits,
+            step,
+        ):
+            x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(x_rec[0])
+            x_rec_img.save(f"{step}.jpg")
+
+        gen_img_list, z_logits_list = taming_decoder.generate_from_prompt(
+            prompt=prompt,
+            lr=lr,
+            target_img_size=target_img_size,
+            num_steps=num_steps,
+            num_augmentations=num_augmentations,
+            init_img_path=init_img_path,
+            loss_type=loss_type,
+            loss_clip_value=loss_clip_value,
+            generation_cb=save_img_cb,
+        )
+
+        init_latent = z_logits_list[-1]
+        torch.save(init_latent, f'{prompt}_logits.pt')
+    gen_img_list, z_logits_list = taming_decoder.zoom(
+        init_latent=init_latent, )
 
     # _gen_img_list, z_logits_list_ = taming_decoder.generate_from_prompt(
     #     prompt="Pokemon of type grass",
